@@ -1,6 +1,70 @@
 '''
-Quick start
------------
+NDB is a high level network management module. IT allows to manage interfaces,
+routes, addresses etc. of connected systems, containers and network
+namespaces.
+
+.. note:: See also the module choice guide: :ref:`choice`
+
+In a nutshell, NDB collects and aggregates netlink events in an SQL database,
+provides Python objects to reflect the system state, and applies changes back
+to the system. The database expects updates only from the sources, no manual
+SQL updates are expected normally.
+
+.. aafig::
+    :scale: 80
+    :textual:
+
+        +----------------------------------------------------------------+
+      +----------------------------------------------------------------+ |
+    +----------------------------------------------------------------+ | |
+    |                                                                | | |
+    |                              kernel                            | |-+
+    |                                                                |-+
+    +----------------------------------------------------------------+
+            |                      | ^                     | ^
+            | `netlink events`     | |                     | |
+            | `inotify events`     | |                     | |
+            | `...`                | |                     | |
+            v                      v |                     v |
+     +--------------+        +--------------+        +--------------+
+     |     source   |        |     source   |        |     source   |<--\\
+     +--------------+        +--------------+        +--------------+   |
+            |                       |                       |           |
+            |                       |                       |           |
+            \\-----------------------+-----------------------/           |
+                                    |                                   |
+              parsed netlink events | `NDB._event_queue`                |
+                                    |                                   |
+                                    v                                   |
+                        +------------------------+                      |
+                        | `NDB.__dbm__()` thread |                      |
+                        +------------------------+                      |
+                                    |                                   |
+                                    v                                   |
+                     +-----------------------------+                    |
+                     | `NDB.schema.load_netlink()` |                    |
+                     | `NDB.objects.*.load*()`     |                    |
+                     +-----------------------------+                    |
+                                    |                                   |
+                                    v                                   |
+                         +----------------------+                       |
+                         |  SQL database        |                       |
+                         |     `SQLite`         |                       |
+                         |     `PostgreSQL`     |                       |
+                         +----------------------+                       |
+                                    |                                   |
+                                    |                                   |
+                                    V                                   |
+                              +---------------+                         |
+                            +---------------+ |                         |
+                          +---------------+ | |  `RTNL_Object.apply()`  |
+                          | NDB object:   | | |-------------------------/
+                          |  `interface`  | | |
+                          |  `address`    | | |
+                          |  `route`      | |-+
+                          |  `...`        |-+
+                          +---------------+
+
 
 The goal of NDB is to provide an easy access to RTNL info and entities via
 Python objects, like `pyroute2.ndb.objects.interface` (see also:
@@ -151,48 +215,34 @@ Change an interface property::
     #            the context manager's __exit__()
 
 '''
-import gc
-import sys
-import json
-import time
-import errno
 import atexit
-import sqlite3
-import logging
-import logging.handlers
-import threading
-import traceback
 import ctypes
 import ctypes.util
+import logging
+import logging.handlers
+import sys
+import threading
+import time
+import traceback
 from functools import partial
-from collections import OrderedDict
+
 from pr2modules import config
-from pr2modules import cli
 from pr2modules.common import basestring
 from pr2modules.netlink import nlmsg_base
 
 ##
 # NDB stuff
 from . import schema
+from .auth_manager import AuthManager
 from .events import (
     DBMExitException,
-    ShutdownException,
     InvalidateHandlerException,
     RescheduleException,
+    ShutdownException,
 )
 from .messages import cmsg, cmsg_event, cmsg_failed, cmsg_sstart
-from .source import Source, SourceProxy
-from .auth_manager import check_auth
-from .auth_manager import AuthManager
-from .objects import RSLV_DELETE
-from .objects.interface import Interface
-from .objects.interface import Vlan
-from .objects.address import Address
-from .objects.route import Route
-from .objects.neighbour import Neighbour
-from .objects.rule import Rule
-from .objects.netns import NetNS
-from .report import RecordSet, Record
+from .transaction import Transaction
+from .view import SourcesView, View
 
 try:
     from urlparse import urlparse
@@ -204,567 +254,19 @@ try:
 except ImportError:
     import Queue as queue
 
-try:
-    import psycopg2
-except ImportError:
-    psycopg2 = None
-
 log = logging.getLogger(__name__)
-_sql_adapters_lock = threading.Lock()
-_sql_adapters_psycopg2_registered = False
-_sql_adapters_sqlite3_registered = False
 
 
-def target_adapter(value):
-    #
-    # MPLS target adapter for SQLite3
-    #
-    return json.dumps(value)
-
-
-class PostgreSQLAdapter(object):
-    def __init__(self, obj):
-        self.obj = obj
-
-    def getquoted(self):
-        return "'%s'" % json.dumps(self.obj)
-
-
-def register_sqlite3_adapters():
-    global _sql_adapters_lock
-    global _sql_adapters_sqlite3_registered
-    with _sql_adapters_lock:
-        if not _sql_adapters_sqlite3_registered:
-            _sql_adapters_sqlite3_registered = True
-            sqlite3.register_adapter(list, target_adapter)
-            sqlite3.register_adapter(dict, target_adapter)
-
-
-def regsiter_postgres_adapters():
-    global _sql_adapters_lock
-    global _sql_adapters_psycopg2_registered
-    with _sql_adapters_lock:
-        if psycopg2 is not None and not _sql_adapters_psycopg2_registered:
-            _sql_adapters_psycopg2_registered = True
-            psycopg2.extensions.register_adapter(list, PostgreSQLAdapter)
-            psycopg2.extensions.register_adapter(dict, PostgreSQLAdapter)
-
-
-class Transaction(object):
-    def __init__(self, log):
-        self.queue = []
-        self.event = threading.Event()
-        self.event.clear()
-        self.log = log.channel('transaction.%s' % id(self))
-        self.log.debug('begin transaction')
-
-    def push(self, *argv):
-        for obj in argv:
-            self.log.debug('queue %s' % type(obj))
-            self.queue.append(obj)
-        return self
-
-    def append(self, obj):
-        self.log.debug('queue %s' % type(obj))
-        self.push(obj)
-        return self
-
-    def pop(self, index=-1):
-        self.log.debug('pop %s' % index)
-        self.queue.pop(index)
-        return self
-
-    def insert(self, index, obj):
-        self.log.debug('insert %i %s' % (index, type(obj)))
-        self.queue.insert(index, obj)
-        return self
-
-    def cancel(self):
-        self.log.debug('cancel transaction')
-        self.queue = []
-        return self
-
-    def wait(self):
-        return self.event.wait()
-
-    def done(self):
-        return self.event.is_set()
-
-    def commit(self):
-        self.log.debug('commit')
-        rollbacks = []
-        for obj in self.queue:
-            rollbacks.append(obj)
-            try:
-                obj.commit()
-            except Exception:
-                for rb in reversed(rollbacks):
-                    try:
-                        rb.rollback()
-                    except Exception as e:
-                        self.log.warning('ignore rollback exception: %s' % e)
-                raise
-        self.event.set()
-        return self
-
-
-class View(dict):
-    '''
-    The View() object returns RTNL objects on demand::
-
-        ifobj1 = ndb.interfaces['eth0']
-        ifobj2 = ndb.interfaces['eth0']
-        # ifobj1 != ifobj2
-    '''
-
-    def __init__(self, ndb, table, chain=None, auth_managers=None):
-        self.ndb = ndb
-        self.log = ndb.log.channel('view.%s' % table)
-        self.table = table
-        self.event = table  # FIXME
-        self.chain = chain
-        self.cache = {}
-        if auth_managers is None:
-            auth_managers = []
-        if chain:
-            auth_managers += chain.auth_managers
-        self.auth_managers = auth_managers
-        self.constraints = {}
-        self.classes = OrderedDict()
-        self.classes['interfaces'] = Interface
-        self.classes['addresses'] = Address
-        self.classes['neighbours'] = Neighbour
-        self.classes['routes'] = Route
-        self.classes['rules'] = Rule
-        self.classes['netns'] = NetNS
-        self.classes['vlans'] = Vlan
-
-    def __enter__(self):
-        return self
-
-    def __exit__(self, exc_type, exc_value, traceback):
-        pass
-
-    @property
-    def default_target(self):
-        if self.table == 'netns':
-            return self.ndb.nsmanager
-        else:
-            return self.ndb.localhost
-
-    @property
-    def context(self):
-        if self.chain is not None:
-            return self.chain.context
-        else:
-            return {}
-
-    def getmany(self, spec, table=None):
-        return self.ndb.schema.get(table or self.table, spec)
-
-    def getone(self, spec, table=None):
-        for obj in self.getmany(spec, table):
-            return obj
-
-    def get(self, spec, table=None):
-        try:
-            return self.__getitem__(spec, table)
-        except KeyError:
-            return None
-
-    def template(self, key, table=None):
-        if self.chain:
-            context = self.chain.context
-        else:
-            context = {}
-        iclass = self.classes[table or self.table]
-
-        spec = (
-            iclass.new_spec(key, self.default_target)
-            .load_context(context)
-            .get_spec
-        )
-
-        return iclass(
-            self,
-            spec,
-            load=False,
-            master=self.chain,
-            auth_managers=self.auth_managers,
-        )
-
-    @cli.change_pointer
-    def create(self, *argspec, **kwspec):
-        iclass = self.classes[self.table]
-
-        if self.chain:
-            context = self.chain.context
-        else:
-            context = {}
-
-        spec = (
-            iclass.new_spec(kwspec or argspec[0], self.default_target)
-            .load_context(context)
-            .get_spec
-        )
-
-        if self.chain:
-            spec['ndb_chain'] = self.chain
-        spec['create'] = True
-        return self[spec]
-
-    @cli.change_pointer
-    def add(self, *argspec, **kwspec):
-        self.log.warning(
-            '''\n
-        The name add() will be removed in future releases, use create()
-        instead. If you believe that the idea to rename is wrong, please
-        file your opinion to the project's bugtracker.
-
-        The reason behind the rename is not to confuse interfaces.add() with
-        bridge and bond port operations, that don't create any new interfaces
-        but work on existing ones.
-        '''
-        )
-        return self.create(*argspec, **kwspec)
-
-    def wait(self, **spec):
-        ret = None
-        timeout = spec.pop('timeout', None)
-        ctime = time.time()
-
-        # install a limited events queue -- for a possible immediate reaction
-        evq = queue.Queue(maxsize=100)
-
-        def handler(evq, target, event):
-            # ignore the "queue full" exception
-            #
-            # if we miss some events here, nothing bad happens: we just
-            # load them from the DB after a timeout, falling back to
-            # the DB polling
-            #
-            # the most important here is not to allocate too much memory
-            try:
-                evq.put_nowait((target, event))
-            except queue.Full:
-                pass
-
-        #
-        hdl = partial(handler, evq)
-        (self.ndb.register_handler(self.ndb.schema.classes[self.event], hdl))
-        #
-        try:
-            ret = self.__getitem__(spec)
-            for key in spec:
-                if ret[key] != spec[key]:
-                    ret = None
-                    break
-        except KeyError:
-            ret = None
-
-        while ret is None:
-            if timeout is not None:
-                if ctime + timeout < time.time():
-                    break
-            try:
-                target, msg = evq.get(timeout=1)
-            except queue.Empty:
-                try:
-                    ret = self.__getitem__(spec)
-                    for key in spec:
-                        if ret[key] != spec[key]:
-                            ret = None
-                            raise KeyError()
-                    break
-                except KeyError:
-                    continue
-
-            #
-            for key, value in spec.items():
-                if key == 'target' and value != target:
-                    break
-                elif value not in (
-                    msg.get(key),
-                    msg.get_attr(msg.name2nla(key)),
-                ):
-                    break
-            else:
-                while ret is None:
-                    try:
-                        ret = self.__getitem__(spec)
-                    except KeyError:
-                        time.sleep(0.1)
-
-        #
-        (self.ndb.unregister_handler(self.ndb.schema.classes[self.event], hdl))
-
-        del evq
-        del hdl
-        gc.collect()
-        if ret is None:
-            raise TimeoutError()
-        return ret
-
-    @check_auth('obj:read')
-    def locate(self, spec=None, table=None, **kwarg):
-        '''
-        This method works like `__getitem__()`, but the important
-        difference is that it uses only key fields to locate the
-        object in the DB, ignoring all other keys.
-
-        It is useful to locate objects that may change attributes
-        during request, like an interface may come up/down, or an
-        address may become primary/secondary, so plain
-        `__getitem__()` will not match while the object still
-        exists.
-        '''
-        if isinstance(spec, Record):
-            spec = spec._as_dict()
-        spec = spec or kwarg
-        if not spec:
-            raise TypeError('got an empty spec')
-
-        table = table or self.table
-        iclass = self.classes[table]
-        spec = iclass.spec_normalize(spec)
-        kspec = self.ndb.schema.compiled[table]['norm_idx']
-
-        request = {}
-        for name in kspec:
-            name = iclass.nla2name(name)
-            if name in spec:
-                request[name] = spec[name]
-
-        if not request:
-            raise KeyError('got an empty key')
-
-        return self[request]
-
-    @check_auth('obj:read')
-    def __getitem__(self, key, table=None):
-
-        ret = self.template(key, table)
-
-        # rtnl_object.key() returns a dcitionary that can not
-        # be used as a cache key. Create here a tuple from it.
-        # The key order guaranteed by the dictionary.
-        cache_key = tuple(ret.key.items())
-
-        rtime = time.time()
-
-        # Iterate all the cache to remove unused and clean
-        # (without any started transaction) objects.
-        for ckey in tuple(self.cache):
-            # Skip the current cache_key to avoid extra
-            # cache del/add records in the logs
-            if ckey == cache_key:
-                continue
-            # The number of referrers must be > 1, the first
-            # one is the cache itself
-            rcount = len(gc.get_referrers(self.cache[ckey]))
-            # Remove only expired items
-            expired = (rtime - self.cache[ckey].atime) > config.cache_expire
-            # The number of changed rtnl_object fields must
-            # be 0 which means that no transaction is started
-            if rcount == 1 and self.cache[ckey].clean and expired:
-                self.log.debug('cache del %s' % (ckey,))
-                self.cache.pop(ckey, None)
-
-        if cache_key in self.cache:
-            self.log.debug('cache hit %s' % (cache_key,))
-            # Explicitly get rid of the created object
-            del ret
-            # The object from the cache has already
-            # registered callbacks, simply return it
-            ret = self.cache[cache_key]
-            ret.atime = rtime
-            return ret
-        else:
-            # Cache only existing objects
-            if self.exists(key):
-                ret.load_sql()
-                self.log.debug('cache add %s' % (cache_key,))
-                self.cache[cache_key] = ret
-
-        ret.register()
-        return ret
-
-    def exists(self, key, table=None):
-        '''
-        Check if the specified object exists in the database::
-
-            ndb.interfaces.exists('eth0')
-            ndb.interfaces.exists({'ifname': 'eth0', 'target': 'localhost'})
-            ndb.addresses.exists('127.0.0.1/8')
-        '''
-        if self.chain:
-            context = self.chain.context
-        else:
-            context = {}
-
-        iclass = self.classes[self.table]
-        key = (
-            iclass.new_spec(key, self.default_target)
-            .load_context(context)
-            .get_spec
-        )
-
-        iclass.resolve(
-            view=self,
-            spec=key,
-            fields=iclass.resolve_fields,
-            policy=RSLV_DELETE,
-        )
-
-        table = table or self.table
-        schema = self.ndb.schema
-        names = schema.compiled[self.table]['all_names']
-
-        self.log.debug('check if the key %s exists in table %s' % (key, table))
-        keys = []
-        values = []
-        for name, value in key.items():
-            nla_name = iclass.name2nla(name)
-            if nla_name in names:
-                name = nla_name
-            if value is not None and name in names:
-                keys.append('f_%s = %s' % (name, schema.plch))
-                values.append(value)
-        spec = schema.fetchone(
-            'SELECT * FROM %s WHERE %s' % (self.table, ' AND '.join(keys)),
-            values,
-        )
-        if spec is not None:
-            self.log.debug('exists')
-            return True
-        else:
-            self.log.debug('not exists')
-            return False
-
-    def __setitem__(self, key, value):
-        raise NotImplementedError()
-
-    def __delitem__(self, key):
-        raise NotImplementedError()
-
-    def __iter__(self):
-        return self.keys()
-
-    def __contains__(self, key):
-        return key in self.dump()
-
-    @check_auth('obj:list')
-    def keys(self):
-        for record in self.dump():
-            yield record
-
-    @check_auth('obj:list')
-    def values(self):
-        for key in self.keys():
-            yield self[key]
-
-    @check_auth('obj:list')
-    def items(self):
-        for key in self.keys():
-            yield (key, self[key])
-
-    @cli.show_result
-    def count(self):
-        return (
-            self.ndb.schema.fetchone('SELECT count(*) FROM %s' % self.table)
-        )[0]
-
-    def __len__(self):
-        return self.count()
-
-    def _keys(self, iclass):
-        return ['target', 'tflags'] + self.ndb.schema.compiled[
-            iclass.view or iclass.table
-        ]['names']
-
-    def _native(self, dump):
-        fnames = next(dump)
-        for record in dump:
-            yield Record(fnames, record, self.classes[self.table])
-
-    @cli.show_result
-    @check_auth('obj:list')
-    def dump(self):
-        iclass = self.classes[self.table]
-        return RecordSet(self._native(iclass.dump(self)))
-
-    @cli.show_result
-    @check_auth('obj:list')
-    def summary(self):
-        iclass = self.classes[self.table]
-        return RecordSet(self._native(iclass.summary(self)))
-
-
-class SourcesView(View):
-    def __init__(self, ndb, auth_managers=None):
-        super(SourcesView, self).__init__(ndb, 'sources')
-        self.classes['sources'] = Source
-        self.cache = {}
-        self.proxy = {}
-        self.lock = threading.Lock()
-        if auth_managers is None:
-            auth_managers = []
-        self.auth_managers = auth_managers
-
-    def async_add(self, **spec):
-        spec = dict(Source.defaults(spec))
-        self.cache[spec['target']] = Source(self.ndb, **spec).start()
-        return self.cache[spec['target']]
-
-    def add(self, **spec):
-        spec = dict(Source.defaults(spec))
-        if 'event' not in spec:
-            sync = True
-            spec['event'] = threading.Event()
-        else:
-            sync = False
-        self.cache[spec['target']] = Source(self.ndb, **spec).start()
-        if sync:
-            self.cache[spec['target']].event.wait()
-        return self.cache[spec['target']]
-
-    def remove(self, target, code=errno.ECONNRESET, sync=True):
-        with self.lock:
-            if target in self.cache:
-                source = self.cache[target]
-                source.close(code=code, sync=sync)
-                return self.cache.pop(target)
-
-    @check_auth('obj:list')
-    def keys(self):
-        for key in self.cache:
-            yield key
-
-    def _keys(self, iclass):
-        return ['target', 'kind']
-
-    def wait(self, **spec):
-        raise NotImplementedError()
-
-    def _summary(self, *argv, **kwarg):
-        return self._dump(*argv, **kwarg)
-
-    def __getitem__(self, key, table=None):
-        if isinstance(key, basestring):
-            target = key
-        elif isinstance(key, dict) and 'target' in key.keys():
-            target = key['target']
-        else:
-            raise KeyError()
-
-        if target in self.cache:
-            return self.cache[target]
-        elif target in self.proxy:
-            return self.proxy[target]
-        else:
-            proxy = SourceProxy(self.ndb, target)
-            self.proxy[target] = proxy
-            return proxy
+NDB_VIEWS_SPECS = (
+    ('interfaces', 'interfaces'),
+    ('addresses', 'addresses'),
+    ('routes', 'routes'),
+    ('neighbours', 'neighbours'),
+    ('af_bridge_fdb', 'fdb'),
+    ('rules', 'rules'),
+    ('netns', 'netns'),
+    ('af_bridge_vlans', 'vlans'),
+)
 
 
 class Log(object):
@@ -861,18 +363,6 @@ class Log(object):
         return self.main.critical(*argv, **kwarg)
 
 
-class ReadOnly(object):
-    def __init__(self, ndb):
-        self.ndb = ndb
-
-    def __enter__(self):
-        self.ndb.schema.allow_write(False)
-        return self
-
-    def __exit__(self, exc_type, exc_value, traceback):
-        self.ndb.schema.allow_write(True)
-
-
 class DeadEnd(object):
     def put(self, *argv, **kwarg):
         raise ShutdownException('shutdown in progress')
@@ -910,17 +400,9 @@ class AuthProxy(object):
         self._ndb = ndb
         self._auth_managers = auth_managers
 
-        for spec in (
-            'interfaces',
-            'addresses',
-            'routes',
-            'neighbours',
-            'rules',
-            'netns',
-            'vlans',
-        ):
-            view = View(self._ndb, spec, auth_managers=self._auth_managers)
-            setattr(self, spec, view)
+        for vtable, vname in NDB_VIEWS_SPECS:
+            view = View(self._ndb, vtable, auth_managers=self._auth_managers)
+            setattr(self, vname, view)
 
 
 class NDB(object):
@@ -945,11 +427,6 @@ class NDB(object):
         if db_provider == 'postgres':
             db_provider = 'psycopg2'
 
-        if db_provider == 'sqlite3':
-            register_sqlite3_adapters()
-        elif db_provider == 'psycopg2':
-            regsiter_postgres_adapters()
-
         self.localhost = localhost
         self.ctime = self.gctime = time.time()
         self.schema = None
@@ -958,7 +435,6 @@ class NDB(object):
             ctypes.util.find_library('c'), use_errno=True
         )
         self.log = Log(log_id=id(self))
-        self.readonly = ReadOnly(self)
         self._auto_netns = auto_netns
         self._db = None
         self._dbm_thread = None
@@ -1026,17 +502,9 @@ class NDB(object):
         for event in tuple(self._dbm_autoload):
             event.wait()
         self._dbm_autoload = None
-        for spec in (
-            'interfaces',
-            'addresses',
-            'routes',
-            'neighbours',
-            'rules',
-            'netns',
-            'vlans',
-        ):
-            view = View(self, spec, auth_managers=[am])
-            setattr(self, spec, view)
+        for vtable, vname in NDB_VIEWS_SPECS:
+            view = View(self, vtable, auth_managers=[am])
+            setattr(self, vname, view)
         # self.query = Query(self.schema)
 
     def _get_view(self, name, chain=None):
@@ -1049,7 +517,15 @@ class NDB(object):
         self.close()
 
     def begin(self):
-        return Transaction(self.log)
+        return Transaction(self.log.channel('transaction'))
+
+    def readonly(self):
+        return self.auth_proxy(
+            AuthManager(
+                {'obj:list': True, 'obj:read': True, 'obj:modify': False},
+                self.log.channel('auth'),
+            )
+        )
 
     def auth_proxy(self, auth_manager):
         return AuthProxy(self, [auth_manager])
@@ -1085,6 +561,9 @@ class NDB(object):
             # shutdown the logger -- free the resources
             self.log.close()
 
+    def backup(self, spec):
+        self.schema.backup(spec)
+
     def reload(self, kinds=None):
         for source in self.sources.values():
             if kinds is not None and source.kind in kinds:
@@ -1138,7 +617,7 @@ class NDB(object):
         def default_handler(target, event):
             if isinstance(getattr(event, 'payload', None), Exception):
                 raise event.payload
-            log.warning('unsupported event ignored: %s' % type(event))
+            log.debug('unsupported event ignored: %s' % type(event))
 
         def check_sources_started(self, _locals, target, event):
             _locals['countdown'] -= 1
@@ -1158,19 +637,18 @@ class NDB(object):
         event_queue = self._event_queue
 
         try:
-            with _sql_adapters_lock:
-                if self._db_provider == 'sqlite3':
-                    self._db = sqlite3.connect(self._db_spec)
-                elif self._db_provider == 'psycopg2':
-                    self._db = psycopg2.connect(**self._db_spec)
-
-            self.schema = schema.init(
+            dbconfig = schema.DBConfig()
+            dbconfig.provider = schema.DBProvider(self._db_provider)
+            dbconfig.spec = self._db_spec
+            self.schema = schema.DBSchema(
+                dbconfig,
                 self,
-                self._db,
-                self._db_provider,
+                self._event_queue,
+                self._event_map,
                 self._db_rtnl_log,
-                id(threading.current_thread()),
+                self.log.channel('schema'),
             )
+
         except Exception as e:
             self._dbm_error = e
             self._dbm_ready.set()
