@@ -1,11 +1,21 @@
 '''
-Structure
-=========
+General structure
+=================
 
 The NDB objects are dictionary-like structures that represent network
-objects -- interfaces, routes, addresses etc. In addition to the
-usual dictionary API they have some NDB-specific methods, see the
-`RTNL_Object` class description below.
+objects -- interfaces, routes, addresses etc. They support the common
+dict API, like item getting, setting, iteration through key and values.
+In addition to that, NDB object add specific calls, see the API section
+below.
+
+Most of the NDB object types store all attributes in a flat one level
+dictionary. Some, like multihop routes, implement nested structures.
+In addition to that, some objects like `Interface` provide views on the
+DB that list only related objects -- addresses, routes and neighbours.
+More on these topic see in the corresponding sections.
+
+NDB objects and RTNL API
+========================
 
 The dictionary fields represent RTNL messages fields and NLA names,
 and the objects are used as argument dictionaries to normal `IPRoute`
@@ -14,7 +24,7 @@ the `IPRoute` methods is valid here as well.
 
 See also: :ref:`iproute`
 
-E.g.::
+.. code-block:: python
 
     # create a vlan interface with IPRoute
     with IPRoute() as ipr:
@@ -38,68 +48,26 @@ Slightly simplifying, if a network object doesn't exist, NDB will run
 an RTNL method with "add" argument, if exists -- "set", and to remove
 an object NDB will call the method with "del" argument.
 
-Accessing objects
-=================
-
-NDB objects are grouped into "views":
-
-    * interfaces
-    * addresses
-    * routes
-    * neighbours
-    * rules
-    * netns
-    * ...
-
-Views are dictionary-like objects that accept strings or dict selectors::
-
-    # access eth0
-    ndb.interfaces["eth0"]
-
-    # access eth0 in the netns test01
-    ndb.sources.add(netns="test01")
-    ndb.interfaces[{"target": "test01", "ifname": "eth0"}]
-
-    # access a route to 10.4.0.0/24
-    ndb.routes["10.4.0.0/24"]
-
-    # same with a dict selector
-    ndb.routes[{"dst": "10.4.0.0", "dst_len": 24}]
-
-Objects cache
-=============
-
-NDB create objects on demand, it doesn't create thousands of route objects
-for thousands of routes by default. The object is being created only when
-accessed for the first time, and stays in the cache as long as it has any
-not committed changes. To inspect cached objects, use views' `.cache`::
-
-    >>> ndb.interfaces.cache.keys()
-    [(('target', u'localhost'), ('tflags', 0), ('index', 1)),  # lo
-     (('target', u'localhost'), ('tflags', 0), ('index', 5))]  # eth3
-
-There is no asynchronous cache invalidation, the cache is being cleaned up
-every time when an object is accessed.
-
 API
 ===
 '''
-import json
-import time
-import errno
-import weakref
-import traceback
-import threading
 import collections
+import errno
+import json
+import threading
+import time
+import traceback
+import weakref
 from functools import partial
+
 from pr2modules import cli
 from pr2modules.netlink.exceptions import NetlinkError
-from ..report import Record
-from ..auth_manager import check_auth
-from ..auth_manager import AuthManager
-from ..events import State
-from ..events import InvalidateHandlerException
+from pr2modules.requests.main import RequestProcessor
+
+from ..auth_manager import AuthManager, check_auth
+from ..events import InvalidateHandlerException, State
 from ..messages import cmsg_event
+from ..report import Record
 
 RSLV_IGNORE = 0
 RSLV_RAISE = 1
@@ -108,54 +76,34 @@ RSLV_DELETE = 3
 
 
 def fallback_add(self, idx_req, req):
+    # ignore all set/get for objects with incomplete idx_req
+    if set(idx_req.keys()) != set(self.kspec):
+        self.log.debug('ignore incomplete idx_req in the fallback')
+        return
+    # try to set the object
     (
         self.ndb._event_queue.put(
             self.sources[self['target']].api(self.api, 'set', **req),
             source=self['target'],
         )
     )
+    # try to get the object
     (
         self.ndb._event_queue.put(
             self.sources[self['target']].api(self.api, 'get', **idx_req),
             source=self['target'],
         )
     )
+    # reload the collected data
     self.load_sql()
 
 
-class Spec(object):
-    '''
-    A universal NDB object spec
-    '''
-
-    def __init__(self, iclass, spec, localhost):
-        if isinstance(spec, Record):
-            spec = spec._as_dict()
-        self.spec = spec
-        self.iclass = iclass
-        self.localhost = localhost
-        self.normalize()
-
-    def normalize(self):
-        self.spec = self.iclass.spec_normalize(self.spec)
-        if 'target' not in self.spec:
-            self.spec['target'] = self.localhost
-        return self
-
-    def load_context(self, context):
-        self.spec.update(context)
-        self.normalize()
-        return self
-
-    @property
-    def get_spec(self):
-        return self.spec
-
-
 class RTNL_Object(dict):
     '''
     The common base class for NDB objects -- interfaces, routes, rules
-    addresses etc.
+    addresses etc. Implements common logic for all the classes, like
+    item setting, commit/rollback, RTNL event filters, loading values
+    from the DB backend etc.
     '''
 
     view = None  # (optional) view to load values for the summary etc.
@@ -165,7 +113,8 @@ class RTNL_Object(dict):
     key_extra_fields = []
     hidden_fields = []
     fields_cmp = {}
-    fields_normalize = {}
+    field_filter = object
+    rollback_chain = []
 
     fallback_for = None
     schema = None
@@ -177,6 +126,7 @@ class RTNL_Object(dict):
     reverse_update = None
     _table = None
     _apply_script = None
+    _apply_script_snapshots = []
     _key = None
     _replace = None
     _replace_on_key_change = False
@@ -285,9 +235,9 @@ class RTNL_Object(dict):
             view.ndb.schema.compiled[cls.table]['fnames'],
         )
 
-    @staticmethod
-    def spec_normalize(spec):
-        return spec
+    @classmethod
+    def spec_normalize(cls, processed, spec):
+        return processed
 
     @staticmethod
     def key_load_context(key, context):
@@ -327,10 +277,14 @@ class RTNL_Object(dict):
         self.load_event.set()
         self.load_debug = False
         self.lock = threading.Lock()
+        self.object_data = RequestProcessor(
+            self.field_filter(), context=weakref.proxy(self)
+        )
         self.kspec = self.schema.compiled[self.table]['idx']
         self.knorm = self.schema.compiled[self.table]['norm_idx']
         self.spec = self.schema.compiled[self.table]['all_names']
         self.names = self.schema.compiled[self.table]['norm_names']
+        self.names_count = [self.names.count(x) for x in self.names]
         self.last_save = None
         if self.event_map is None:
             self.event_map = {}
@@ -373,32 +327,39 @@ class RTNL_Object(dict):
         self._init_complete = True
 
     @classmethod
-    def new_spec(cls, spec, localhost):
-        return Spec(cls, spec, localhost)
+    def new_spec(cls, spec, context=None, localhost=None):
+        if isinstance(spec, Record):
+            spec = spec._as_dict()
+        rp = RequestProcessor(cls.field_filter(), context=spec, prime=spec)
+        if isinstance(context, dict):
+            rp.update(context)
+        if 'target' not in rp and localhost is not None:
+            rp['target'] = localhost
+        return cls.spec_normalize(rp, spec)
 
     @staticmethod
     def resolve(view, spec, fields, policy=RSLV_IGNORE):
         '''
-        Resolve specific fields e.g. convert port ifname into index::
+        Resolve specific fields e.g. convert port ifname into index.
         '''
         for field in fields:
-            ref = spec.get(field)
-            if ref is not None and not isinstance(ref, int):
-                try:
-                    spec[field] = view[ref]['index']
-                except KeyError:
-                    if policy == RSLV_RAISE:
-                        raise
-                    elif policy == RSLV_NONE:
-                        spec[field] = None
-                    elif policy == RSLV_DELETE:
-                        del spec[field]
-                    elif policy == RSLV_IGNORE:
-                        pass
-                    else:
-                        raise TypeError('unknown rslv policy')
-
-        return spec
+            reference = spec.get(field)
+            try:
+                if isinstance(reference, dict) and 'index' in reference:
+                    spec[field] = reference['index']
+                elif reference is not None and not isinstance(reference, int):
+                    spec[field] = view[reference]['index']
+            except (KeyError, TypeError):
+                if policy == RSLV_RAISE:
+                    raise
+                elif policy == RSLV_NONE:
+                    spec[field] = None
+                elif policy == RSLV_DELETE:
+                    del spec[field]
+                elif policy == RSLV_IGNORE:
+                    pass
+                else:
+                    raise TypeError('unknown rslv policy')
 
     def mark_tflags(self, mark):
         pass
@@ -442,26 +403,31 @@ class RTNL_Object(dict):
 
     @check_auth('obj:modify')
     def __setitem__(self, key, value):
-        if key in self.fields_normalize:
-            value = self.fields_normalize[key](value)
-
-        if self.state == 'system' and key in self.knorm:
-            if self._replace_on_key_change:
-                self.log.debug('prepare replace for key %s' % (self.key))
-                self._replace = type(self)(
-                    self.view, self.key, auth_managers=self.auth_managers
-                )
-                self.state.set('replace')
-            else:
-                raise ValueError('attempt to change a key field (%s)' % key)
-        if key in ('net_ns_fd', 'net_ns_pid'):
-            self.state.set('setns')
-        if value != self.get(key, None):
-            if key != 'target':
-                self.changed.add(key)
-            dict.__setitem__(self, key, value)
+        for nkey, nvalue in self.object_data.filter(key, value).items():
+            if self.get(nkey) == nvalue:
+                continue
+            if self.state == 'system' and nkey in self.knorm:
+                if self._replace_on_key_change:
+                    self.log.debug(
+                        f'prepare replace {nkey} = {nvalue} in {self.key}'
+                    )
+                    self._replace = type(self)(
+                        self.view, self.key, auth_managers=self.auth_managers
+                    )
+                    self.state.set('replace')
+                else:
+                    raise ValueError(
+                        'attempt to change a key field (%s)' % nkey
+                    )
+            if nkey in ('net_ns_fd', 'net_ns_pid'):
+                self.state.set('setns')
+            if nvalue != self.get(nkey, None):
+                if nkey != 'target':
+                    self.changed.add(nkey)
+                dict.__setitem__(self, nkey, nvalue)
 
     def fields(self, *argv):
+        # TODO: deprecate and move to show()
         Fields = collections.namedtuple('Fields', argv)
         return Fields(*[self[key] for key in argv])
 
@@ -470,13 +436,41 @@ class RTNL_Object(dict):
 
     @cli.change_pointer
     def create(self, **spec):
+        '''
+        Create an RTNL object of the same type, and add it to the
+        commit chain. The spec format depends on the object.
+
+        The method allows to chain creation of multiple objects sharing
+        the same context.
+
+        .. code-block:: python
+
+            (
+                ndb.interfaces['eth0']                     # 1.
+                .set(state="up")                           # 2.
+                .ipaddr                                    # 3.
+                .create(address='10.0.0.1', prefixlen=24)  # 4. <- create()
+                .create(address='10.0.0.2', prefixlen=24)  # 5. <- create()
+                .commit()                                  # 6.
+            )
+
+        Here:
+
+        1. returns an interface object `eth0`
+        2. sets `state="up"` and returns the object itself
+        3. returns an address view, that uses `eth0` as the context
+        4. creates an IP address, the interface lookup is done via context
+        5. creates another IP address -- same type, same context
+        6. commits the changes in the order: (interface `state="up"`;
+           address `10.0.0.1/24`; address `10.0.0.2/24`)
+        '''
         spec['create'] = True
         spec['ndb_chain'] = self
         return self.view[spec]
 
     @cli.show_result
     @check_auth('obj:read')
-    def show(self, **kwarg):
+    def show(self, fmt=None):
         '''
         Return the object in a specified format. The format may be
         specified with the keyword argument `format` or in the
@@ -484,12 +478,7 @@ class RTNL_Object(dict):
 
         TODO: document different formats
         '''
-        fmt = kwarg.pop(
-            'format',
-            kwarg.pop(
-                'fmt', self.view.ndb.config.get('show_format', 'native')
-            ),
-        )
+        fmt = fmt or self.view.ndb.config.get('show_format', 'native')
         if fmt == 'native':
             return dict(self)
         else:
@@ -653,6 +642,7 @@ class RTNL_Object(dict):
             return self.remove().apply()
         else:
             snapshot.state.set(self.state.get())
+            snapshot.rollback_chain = self._apply_script_snapshots
             snapshot.apply(rollback=True)
             for link, snp in snapshot.snapshot_deps:
                 link.rollback(snapshot=snp)
@@ -672,8 +662,8 @@ class RTNL_Object(dict):
     @check_auth('obj:modify')
     def commit(self):
         '''
-        Try to commit the pending changes. If the commit fails,
-        automatically revert the state.
+        Commit the pending changes. If an exception is raised during
+        `commit()`, automatically `rollback()` to the latest saved snapshot.
         '''
         if self.clean:
             return self
@@ -685,8 +675,9 @@ class RTNL_Object(dict):
         if self.state == 'invalid':
             # Save values, try to apply
             save = dict(self)
+            self.last_save = -1
             try:
-                return self.apply()
+                return self.apply(mode='commit')
             except Exception as e_i:
                 # Save the debug info
                 e_i.trace = traceback.format_exc()
@@ -709,9 +700,10 @@ class RTNL_Object(dict):
         # manually by `ndb.schema.purge_snapshots()` -- to invalidate all
         # the snapshots and to drop the associated tables.
 
+        self.last_save = self.snapshot()
         # Apply the changes
         try:
-            self.apply()
+            self.apply(mode='commit')
         except Exception as e_c:
             # Rollback in the case of any error
             try:
@@ -730,6 +722,10 @@ class RTNL_Object(dict):
         return self
 
     def remove(self):
+        '''
+        Set the desired state to `remove`, so the next `apply()` call will
+        delete the object from the system.
+        '''
         with self.lock:
             self.state.set('remove')
             return self
@@ -757,15 +753,6 @@ class RTNL_Object(dict):
         self.log.debug('check: True')
         return True
 
-    def make_prime(self):
-        return dict(
-            [
-                (x, self[self.iclass.nla2name(x)])
-                for x in self.schema.compiled[self.table]['idx']
-                if self.iclass.nla2name(x) in self
-            ]
-        )
-
     def make_req(self, prime):
         req = dict(prime)
         for key in self.changed:
@@ -775,9 +762,6 @@ class RTNL_Object(dict):
     def make_idx_req(self, prime):
         return prime
 
-    def sync_req(self, prime):
-        return prime
-
     def get_count(self):
         conditions = []
         values = []
@@ -798,10 +782,27 @@ class RTNL_Object(dict):
         pass
 
     @check_auth('obj:modify')
-    def apply(self, rollback=False, req_filter=None):
+    def save_context(self):
+        if self.state == 'invalid':
+            self.last_save = -1
+        else:
+            self.last_save = self.snapshot()
+        return self
+
+    @check_auth('obj:modify')
+    def apply(self, rollback=False, req_filter=None, mode='apply'):
         '''
-        Create a snapshot and apply pending changes. Do not revert
-        the changes in the case of an exception.
+        Apply the pending changes. If an exception is raised during
+        `apply()`, no `rollback()` is called. No automatic snapshots
+        are madre.
+
+        In order to properly revert the changes, you have to run::
+
+            obj.save_context()
+            try:
+                obj.apply()
+            except Exception:
+                obj.rollback()
         '''
 
         # Resolve the fields
@@ -812,13 +813,6 @@ class RTNL_Object(dict):
             policy=RSLV_RAISE,
         )
 
-        # Save the context
-        if not rollback:
-            if self.state == 'invalid':
-                self.last_save = -1
-            else:
-                self.last_save = self.snapshot()
-
         self.log.debug('events log: %s' % str(self.state.events))
         self.log.debug('run apply')
         self.load_event.clear()
@@ -836,8 +830,12 @@ class RTNL_Object(dict):
             state = self.state.get()
 
         # Create the request.
-        prime = self.make_prime()
-        req = self.sync_req(self.make_req(prime))
+        prime = {
+            x: self[x]
+            for x in self.schema.compiled[self.table]['norm_idx']
+            if self.get(x) is not None
+        }
+        req = self.make_req(prime)
         idx_req = self.make_idx_req(prime)
         self.log.debug('apply req: %s' % str(req))
         self.log.debug('apply idx_req: %s' % str(idx_req))
@@ -848,10 +846,8 @@ class RTNL_Object(dict):
                 if k not in req and v is not None:
                     req[k] = v
             if self.master is not None:
-                req = (
-                    self.new_spec(req, self.ndb.localhost)
-                    .load_context(self.master.context)
-                    .get_spec
+                req = self.new_spec(
+                    req, self.master.context, self.ndb.localhost
                 )
 
             method = 'add'
@@ -980,12 +976,16 @@ class RTNL_Object(dict):
                         obj.apply()
                     except Exception as e:
                         self.errors.append((time.time(), obj, e))
+            for obj in reversed(self.rollback_chain):
+                obj.rollback()
         else:
             apply_script = self._apply_script
             self._apply_script = []
-            for op, argv, kwarg in apply_script:
+            for op, kwarg in apply_script:
+                kwarg['self'] = self
+                kwarg['mode'] = mode
                 ret = self.use_db_resync(
-                    lambda x: not isinstance(x, KeyError), op, *argv, **kwarg
+                    lambda x: not isinstance(x, KeyError), op, tuple(), kwarg
                 )
                 if not isinstance(ret, list):
                     ret = [ret]
@@ -996,8 +996,10 @@ class RTNL_Object(dict):
                         self._apply_script_snapshots.append(obj)
         return self
 
-    def use_db_resync(self, criteria, method, *argv, **kwarg):
+    def use_db_resync(self, criteria, method, argv=None, kwarg=None):
         ret = None
+        argv = argv or []
+        kwarg = kwarg or {}
         self.log.debug(f'criteria {criteria}')
         self.log.debug(f'method {method}, {argv}, {kwarg}')
         for attempt in range(3):
@@ -1029,6 +1031,18 @@ class RTNL_Object(dict):
         for key, value in data.items():
             self.load_value(key, value)
 
+    def update_from_sql(self, spec):
+        '''
+        A bit special case: we can have several fields with
+        non unique names.
+        '''
+        for key, count, value in zip(self.names, self.names_count, spec):
+            if count == 1 or value is not None:
+                self.load_value(key, value)
+
+    def load_direct(self, key, value):
+        super(RTNL_Object, self).__setitem__(key, value)
+
     def load_value(self, key, value):
         '''
         Load a value and clean up the `self.changed` set if the
@@ -1073,7 +1087,8 @@ class RTNL_Object(dict):
         spec = self.ndb.schema.fetchone(
             'SELECT * FROM %s WHERE %s' % (table, ' AND '.join(keys)), values
         )
-        self.log.debug('load_sql: %s' % str(spec))
+        self.log.debug('load_sql load: %s' % str(spec))
+        self.log.debug('load_sql names: %s' % str(self.names))
         if set_state:
             with self.lock:
                 if spec is None:
@@ -1082,7 +1097,7 @@ class RTNL_Object(dict):
                         self.state.set('invalid')
                         self.changed = set()
                 elif self.state not in ('remove', 'setns'):
-                    self.update(dict(zip(self.names, spec)))
+                    self.update_from_sql(spec)
                     self.state.set('system')
         return spec
 
@@ -1095,14 +1110,14 @@ class RTNL_Object(dict):
         # ...
 
         # full match
-        for name in self.knorm:
-            value = self.get(name)
+        for norm, name in zip(self.knorm, self.kspec):
+            value = self.get(norm)
             if name == 'target':
                 if value != target:
                     return
             elif name == 'tflags':
                 continue
-            elif value != (event.get_attr(name) or event.get(name)):
+            elif value not in (event.get_attr(name), event.get(norm)):
                 return
 
         self.log.debug('load_rtnl: %s' % str(event.get('header')))
